import json
import logging
from typing import ClassVar

import gevent
from markupsafe import escape
from mxcubecore import HardwareRepository as HWR
from mxcubecore import queue_entry
from mxcubecore.HardwareObjects.abstract.AbstractSampleChanger import (
    SampleChanger,
    SampleChangerState,
)

from mxcubeweb.core.adapter.adapter_base import AdapterBase
from mxcubeweb.core.models.adaptermodels import (
    SampleChangerCommandInputModel,
    SampleInputModel,
)
from mxcubeweb.core.models.configmodels import ResourceHandlerConfigModel

logger = logging.getLogger("MX3.HWR")

resource_handler_config = ResourceHandlerConfigModel(
    name="sample_changer",
    url_prefix="/mxcube/api/v0.1/sample_changer",
    attributes=[
        "loaded_sample",
        "get_contents",
        "get_value",
        "sync_with_crims",
    ],
    commands=[
        "select_location",
        "scan_location",
        "unmount_current",
        "mount_sample",
        "send_command",
    ],
)


class SampleChangerAdapter(AdapterBase):
    """Adapter for AbstractSampleChanger."""

    SUPPORTED_TYPES: ClassVar[list[object]] = [SampleChanger]

    def __init__(self, ho, role, app):
        super().__init__(ho, role, app, resource_handler_config)
        self.app = app

        sc = HWR.beamline.sample_changer

        # Initialize hwobj signals
        sc.connect("stateChanged", self._sc_state_changed)
        sc.connect("isCollisionSafe", self._is_collision_safe)
        sc.connect("loadedSampleChanged", self._loaded_sample_changed)
        sc.connect("contentsUpdated", self._sc_contents_update)

        if HWR.beamline.sample_changer_maintenance is not None:
            HWR.beamline.sample_changer_maintenance.connect(
                "globalStateChanged", self._sc_maintenance_update
            )

            HWR.beamline.sample_changer_maintenance.connect(
                "gripperChanged", self._gripper_changed
            )

    def _sc_state_changed(self, *args):
        new_state = args[0]
        state_str = SampleChangerState.STATE_DESC.get(new_state, "Unknown").upper()
        self.app.server.emit("sc_state", state_str, namespace="/hwr")

    def _is_collision_safe(self, *args):
        new_state = args[0]
        # we are only interested when it becames true
        if new_state:
            msg = {
                "signal": "isCollisionSafe",
                "message": "Sample moved to safe area",
            }
            self.app.server.emit("sc", msg, namespace="/hwr")

    def _loaded_sample_changed(self, sample):
        if hasattr(sample, "get_address"):
            address = sample.get_address()
            barcode = sample.get_id()
        else:
            address = ""
            barcode = ""

        logging.getLogger("HWR").info("Loaded sample changed: %s", address)

        try:
            sample_id = address

            if HWR.beamline.sample_changer.has_loaded_sample():
                self.app.lims.set_current_sample(sample_id)
            else:
                sample = HWR.beamline.sample_changer.get_loaded_sample()
                address = sample.get_address() if sample else None
                self.app.lims.set_current_sample(address)

            self.app.server.emit(
                "loaded_sample_changed",
                {"address": address, "barcode": barcode},
                namespace="/hwr",
            )

            self._sc_load_ready(address)
        except Exception:
            logging.getLogger("HWR").exception("Error setting loaded sample")

    def _sc_load_ready(self, location):
        msg = {
            "signal": "loadReady",
            "location": location,
            "message": "Sample changer, loaded sample",
        }

        self.app.server.emit("sc", msg, namespace="/hwr")

    def _sc_unload(self, location):
        msg = {
            "signal": "operatingSampleChanger",
            "location": location,
            "message": "Please wait, unloading sample",
        }

        self.app.server.emit("sc", msg, namespace="/hwr")

    def _sc_contents_update(self):
        self.app.server.emit("sc_contents_update", {}, namespace="/hwr")

    def _sc_maintenance_update(self, *args):
        if len(args) == 3:
            # Restore the `global_state` parameter removed in this commit 337efd37
            global_state, cmd_state, message = args
        else:
            # Be backward compatible with HW objects which are emitting signal with
            # 2 arguments
            global_state = {}
            cmd_state, message = args

        try:
            self.app.server.emit(
                "sc_maintenance_update",
                {
                    "global_state": global_state,
                    "commands_state": json.dumps(cmd_state),
                    "message": message,
                },
                namespace="/hwr",
            )
        except Exception:
            logging.getLogger("HWR").exception("error sending message")

    def _gripper_changed(self):
        self.app.queue.queue_clear()
        self.app.server.emit(
            "queue", {"Signal": "update", "message": "all"}, namespace="/hwr"
        )

    def _mount_sample(self, sample: SampleInputModel):
        sc = HWR.beamline.sample_changer
        res = False

        try:
            msg = {
                "signal": "operatingSampleChanger",
                "location": sample.location,
                "message": "Please wait, loading sample",
            }

            self.app.server.emit("sc", msg, namespace="/hwr")

            sid = self.app.lims.get_current_sample().get("sampleID", False)
            current_queue = self.app.queue.queue_to_dict()

            if sample.location != "Manual":
                msg = f"Mounting sample: {sample.location} ({sample.sample_name})"
                logging.getLogger("user_level_log").info(msg)

                if (
                    not sc.get_loaded_sample()
                    or sc.get_loaded_sample().get_address() != sample.location
                ):
                    res = sc.load(sample.sample_id, wait=True)

                if (
                    res
                    and HWR.beamline.queue_manager.centring_method
                    == queue_entry.CENTRING_METHOD.LOOP
                    and not HWR.beamline.diffractometer.in_plate_mode()
                    and not self.app.harvester.mount_from_harvester()
                ):
                    HWR.beamline.diffractometer.reject_centring()
                    msg = "Starting autoloop centring ..."
                    logging.getLogger("MX3.HWR").info(msg)
                    HWR.beamline.diffractometer.start_centring_method(
                        HWR.beamline.diffractometer.C3D_MODE
                    )
                elif HWR.beamline.diffractometer.in_plate_mode():
                    msg = "Starting autoloop Focusing ..."
                    logging.getLogger("MX3.HWR").info(msg)
                    sc.move_to_crystal_position(None)

            else:
                msg = f"Mounting sample: {sample.sample_name}"
                logging.getLogger("user_level_log").info(msg)

                self.app.lims.set_current_sample(sample.sample_id)
                res = True

        except Exception as ex:
            logging.getLogger("MX3.HWR").exception("[SC] sample could not be mounted")
            raise RuntimeError(str(ex)) from ex
        else:
            # Clean up if the new sample was mounted or the current sample was
            # unmounted and the new one, for some reason, failed to mount
            if res or (not res and not sc.get_loaded_sample()):
                HWR.beamline.sample_view.clear_all()

                # We remove the current sample from the queue, if we are moving
                # from one sample to another and the current sample is in the queue

                if sid and current_queue.get(sid, False):
                    node_id = current_queue[sid]["queueID"]
                    self.app.queue.set_enabled_entry(node_id, False)  # noqa: FBT003
                    self.app.queue.queue_toggle_sample(
                        self.app.queue.get_entry(node_id)[1]
                    )
        finally:
            self._sc_load_ready(sample.location)

        return res

    def _unmount_sample(self, location):
        try:
            self._sc_unload(location)

            if location != "Manual":
                HWR.beamline.sample_changer.unload(location, wait=False)
            else:
                self.app.lims.set_current_sample(None)
                self._sc_load_ready(location)

            msg = f"[SC] unmounted {location}"
            logging.getLogger("MX3.HWR").info(msg)
        except Exception:
            msg = "[SC] sample could not be mounted"
            logging.getLogger("MX3.HWR").exception(msg)
            raise
        else:
            HWR.beamline.queue_model.mounted_sample = ""
            HWR.beamline.sample_view.clear_all()

    def get_value(self) -> dict:
        if HWR.beamline.sample_changer_maintenance is not None:
            global_state, cmdstate, msg = self.get_global_state()

            cmds = HWR.beamline.sample_changer_maintenance.get_cmd_info()

        else:
            global_state = {}
            cmdstate = "SC maintenance controller not defined"
            cmds = []
            msg = ""

        contents = self._ho.get_contents_as_dict()
        address, barcode = self.get_loaded_sample()

        loaded_sample = {"address": address, "barcode": barcode}

        try:
            state = HWR.beamline.sample_changer.get_status().upper()
        except Exception:
            logging.getLogger("MX3.HWR").exception("")
            state = "OFFLINE"

        return {
            "state": state,
            "loaded_sample": loaded_sample,
            "contents": contents,
            "global_state": {
                "global_state": global_state,
                "commands_state": cmdstate,
            },
            "cmds": {"cmds": cmds},
            "msg": msg,
            "plate_mode": HWR.beamline.diffractometer.in_plate_mode(),
        }

    def state(self):
        return "READY" if self._ho.is_ready() else "BUSY"

    def loaded_sample(self) -> dict:
        if self._ho.has_loaded_sample():
            address, barcode = self._ho.get_loaded_sample()
        else:
            address, barcode = "", ""

        return {"address": address, "barcode": barcode}

    def get_contents(self):
        return self._ho.get_contents_as_dict()

    def get_loaded_sample(self):
        try:
            sample = HWR.beamline.sample_changer.get_loaded_sample()
        except Exception:
            logging.getLogger("MX3.HWR").exception("")
            sample = None

        if sample is not None:
            address = sample.get_address()
            barcode = sample.get_id()
        else:
            address = ""
            barcode = ""

        return address, barcode

    def get_global_state(self):
        try:
            return HWR.beamline.sample_changer_maintenance.get_global_state()
        except Exception:
            logging.getLogger("MX3.HWR").exception("Could not get sc global state")
            return (
                {},
                "SC maintenance controller doesn't response",
                "Can't retrieve the global state",
            )

    def select_location(self, loc: str):
        self._ho.select(loc)
        return self._ho.get_contents_as_dict()

    def scan_location(self, loc: str):
        loc = None if loc == "" else loc
        self._ho.scan(loc, recursive=True)
        return self._ho.get_contents_as_dict()

    def mount_sample(self, sample: SampleInputModel, wait=True):  # noqa: FBT002
        if wait:
            self._mount_sample(sample)
        else:
            gevent.spawn(self._mount_sample, sample)

        return HWR.beamline.sample_changer.get_contents_as_dict()

    def unmount_current(self):
        sc_sample = HWR.beamline.sample_changer.get_loaded_sample()
        if sc_sample:
            location = sc_sample.get_address()
            self._unmount_sample(location)
        else:
            self._unmount_sample("Manual")

        return HWR.beamline.sample_changer.get_contents_as_dict()

    def send_command(self, command: SampleChangerCommandInputModel):
        try:
            return {
                "response": HWR.beamline.sample_changer_maintenance.send_command(
                    command.cmd, command.arguments
                )
            }
        except Exception as _ex:
            logger.exception("SC cannot execute command %s", command.cmd)
            msg = f"Cannot execute command {escape(command.cmd)}"
            raise RuntimeError(msg) from _ex

    def sync_with_crims(self):
        """Synchronize with Crims.

        To be used mostly when diffractometer is in plate mode.
        This returns a list of crystal dicts available in Crims
        that have been harvested.
        With this, the user can visualize more easily
        where the crystals are in the plate GUI.
        """
        xtal_list = []
        try:
            processing_plan = HWR.beamline.sample_changer.sync_with_crims()
            for x in processing_plan.plate.xtal_list:
                response = {
                    "crystal_uuid": x.crystal_uuid,
                    "row": x.row,
                    "column": x.column,
                    "shelf": x.shelf,
                    "offset_x": x.offset_x,
                    "offset_y": x.offset_y,
                    "image_url": x.image_url,
                    "image_date": x.image_date,
                    "sample": x.sample,
                }
                xtal_list.append(response)
        except Exception:
            logging.getLogger("MX3.HWR").exception("Could not get crystal List")

        return {"xtal_list": xtal_list}



import contextlib
import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.Harvester import HarvesterState
from mxcubecore.model import queue_model_objects as qmo

from mxcubeweb.app import MXCUBEApplication as mxcube
from mxcubeweb.core.components.queue import (
    COLLECTED,
    FAILED,
    RUNNING,
    WARNING,
)
from mxcubeweb.core.util.networkutils import RateLimited
from mxcubeweb.server import Server as server


def last_queue_node():
    node = HWR.beamline.queue_manager._current_queue_entries[-1].get_data_model()

    # Reference collections are orphans, the node we want is the
    # characterisation not the reference collection itself
    if "ref" in node.get_name():
        parent = node.get_parent()
        node = parent._children[0]

    res = mxcube.queue.node_index(node)
    res["node"] = node

    return res


centringSignals = [
    "centringInvalid",
    "newAutomaticCentringPoint",
    "centringStarted",
    "centringAccepted",
    "centringMoving",
    "centringFailed",
    "centringSuccessful",
    "centringSnapshots",
]

task_signals = {  # missing egyscan, xrf, etc...
    "collectStarted": "Data collection has started",
    "collectOscillationStarted": "Data collection oscillation has started",
    "collectOscillationFailed": "Data collection oscillacion has failed",
    "collectOscillationFinished": "Data collection oscillacion has finished",
    "collectEnded": "Data collection has finished",
    "warning": "Data collection finished with a warning",
    "collect_finished": "Data collection has finished",
    "collectImageTaken": "Image acquired",
}

motor_signals = {
    "actuatorStateChanged": "actuatorStateChanged",
    "minidiffPhaseChanged": "minidiffPhaseChanged",
    "minidiffTransferModeChanged": "minidiffTransferModeChanged",
    "minidiffSampleIsLoadedChanged": "minidiffSampleIsLoadedChanged",
    "zoomMotorPredefinedPositionChanged": "zoomMotorPredefinedPositionChanged",
}


def handle_auto_mount_next(entry):
    model = entry.get_data_model()

    if isinstance(model.get_parent(), qmo.TaskGroup):
        auto_mount = mxcube.queue.get_auto_mount_sample()
        tgroup = entry.get_data_model()
        tgroup_list = entry.get_data_model().get_parent().get_children()

        try:
            last_gentry = tgroup_list.index(tgroup) == (len(tgroup_list) - 1)
        except ValueError:
            last_gentry = None

        if not auto_mount and last_gentry:
            msg = "Not mounting next sample automatically (Auto mount next)"
            logging.getLogger("user_level_log").info(msg)


def harvester_state_changed(*args):
    new_state = args[0]
    state_str = HarvesterState.STATE_DESC.get(new_state, "Unknown").upper()
    server.emit("harvester_state", state_str, namespace="/hwr")


def harvester_contents_update():
    server.emit("harvester_contents_update")


def get_task_state(entry):
    node_id = entry.get_data_model()._node_id
    _, state = mxcube.queue.get_node_state(node_id)
    node_index = mxcube.queue.node_index(entry.get_data_model())

    return {
        "Signal": "",
        "Message": "",
        "taskIndex": node_index["idx"],
        "queueID": node_id,
        "sample": node_index["sample"],
        "state": state,
        "progress": 1 if state == COLLECTED else 0,
    }


def queue_execution_entry_started(entry, message=None):
    handle_auto_mount_next(entry)

    if not mxcube.queue.is_interleaved(entry.get_data_model()):
        server.emit("task", get_task_state(entry), namespace="/hwr")


def queue_execution_entry_finished(entry, message):
    handle_auto_mount_next(entry)

    if not mxcube.queue.is_interleaved(entry.get_data_model()):
        server.emit("task", get_task_state(entry), namespace="/hwr")

    mxcube.queue.queue_toggle_sample(entry)


def queue_execution_started(entry, queue_state=None):
    state = queue_state if queue_state else mxcube.queue.queue_exec_state()
    msg = {"Signal": state, "Message": "Queue execution started"}

    server.emit("queue", msg, namespace="/hwr")


def queue_execution_finished(entry, queue_state=None):
    state = queue_state if queue_state else mxcube.queue.queue_exec_state()
    msg = {"Signal": state, "Message": "Queue execution stopped"}

    mxcube.queue.enable_sample_entries(mxcube.TEMP_DISABLED, True)
    mxcube.TEMP_DISABLED = []

    server.emit("queue", msg, namespace="/hwr")


def queue_execution_stopped(*args):
    msg = {
        "Signal": "QueueStopped",
        "Message": "Queue execution stopped",
    }

    server.emit("queue", msg, namespace="/hwr")


def queue_execution_paused(state):
    if state:
        msg = {
            "Signal": "QueuePaused",
            "Message": "Queue execution paused",
        }
    else:
        msg = {
            "Signal": "QueueRunning",
            "Message": "Queue execution paused",
        }

    server.emit("queue", msg, namespace="/hwr")


def queue_execution_failed(entry):
    msg = {
        "Signal": mxcube.queue.queue_exec_state(),
        "Message": "Queue execution stopped",
    }

    server.emit("queue", msg, namespace="/hwr")


def collect_oscillation_started(*args):
    node = last_queue_node()

    if not mxcube.queue.is_interleaved(node["node"]):
        msg = {
            "Signal": "collectOscillationStarted",
            "Message": task_signals["collectOscillationStarted"],
            "taskIndex": node["idx"],
            "queueID": node["queue_id"],
            "sample": node["sample"],
            "state": RUNNING,
            "progress": 0,
        }

        logging.getLogger("HWR").debug("[TASK CALLBACK] " + str(msg))

        try:
            server.emit("task", msg, namespace="/hwr")
        except Exception:
            logging.getLogger("HWR").error("error sending message: " + str(msg))


def collect_image_taken(frame):
    try:
        node = last_queue_node()
    except IndexError:
        node = None

    if node and not mxcube.queue.is_interleaved(node["node"]):
        progress = mxcube.queue.get_task_progress(last_queue_node()["node"], frame)

        msg = {
            "Signal": "collectImageTaken",
            "Message": task_signals["collectImageTaken"],
            "taskIndex": node["idx"],
            "queueID": node["queue_id"],
            "sample": node["sample"],
            "state": RUNNING if progress < 1 else COLLECTED,
            "progress": progress,
        }
        try:
            _emit_progress(msg)
        except Exception:
            logging.getLogger("HWR").error("error sending message: " + str(msg))


@RateLimited(1)
def _emit_progress(msg):
    logging.getLogger("HWR").debug("[TASK CALLBACK] " + str(msg))
    server.emit("task", msg, namespace="/hwr")


def collect_oscillation_failed(
    owner=None,
    status=FAILED,
    state=None,
    lims_id="",
    osc_id=None,
    params=None,
):
    node = last_queue_node()

    mxcube.NODE_ID_TO_LIMS_ID[node["queue_id"]] = lims_id

    if not mxcube.queue.is_interleaved(node["node"]):
        with contextlib.suppress(Exception):
            HWR.beamline.get_dc(lims_id)

        msg = {
            "Signal": "collectOscillationFailed",
            "Message": task_signals["collectOscillationFailed"],
            "taskIndex": node["idx"],
            "queueID": node["queue_id"],
            "sample": node["sample"],
            "state": FAILED,
            "progress": 0,
        }

        logging.getLogger("HWR").debug("[TASK CALLBACK] " + str(msg))

        try:
            server.emit("task", msg, namespace="/hwr")
        except Exception:
            logging.getLogger("HWR").error("error sending message: " + str(msg))


def collect_oscillation_finished(owner, status, state, lims_id, osc_id, params):
    node = last_queue_node()
    mxcube.NODE_ID_TO_LIMS_ID[node["queue_id"]] = lims_id

    if not mxcube.queue.is_interleaved(node["node"]):
        msg = {
            "Signal": "collectOscillationFinished",
            "Message": task_signals["collectOscillationFinished"],
            "taskIndex": node["idx"],
            "queueID": node["queue_id"],
            "sample": node["sample"],
            "state": COLLECTED,
            "progress": 1,
        }

        logging.getLogger("HWR").debug("[TASK CALLBACK] " + str(msg))

        try:
            server.emit("task", msg, namespace="/hwr")
        except Exception:
            logging.getLogger("HWR").error("error sending message: " + str(msg))


def collect_ended(owner, success, message):
    node = last_queue_node()

    if not mxcube.queue.is_interleaved(node["node"]):
        state = COLLECTED if success else WARNING

        msg = {
            "Signal": "collectOscillationFinished",
            "Message": message,
            "taskIndex": node["idx"],
            "queueID": node["queue_id"],
            "sample": node["sample"],
            "state": state,
            "progress": 1,
        }

        logging.getLogger("HWR").debug("[TASK CALLBACK] " + str(msg))

        try:
            server.emit("task", msg, namespace="/hwr")
        except Exception:
            logging.getLogger("HWR").error("error sending message: " + str(msg))


def collect_started(*args, **kwargs):
    node = last_queue_node()

    if not mxcube.queue.is_interleaved(node["node"]):
        msg = {
            "Signal": kwargs["signal"],
            "Message": task_signals[kwargs["signal"]],
            "taskIndex": last_queue_node()["idx"],
            "queueID": last_queue_node()["queue_id"],
            "sample": last_queue_node()["sample"],
            "state": RUNNING,
            "progress": 0,
        }

        logging.getLogger("HWR").debug("[TASK CALLBACK] " + str(msg))

        try:
            server.emit("task", msg, namespace="/hwr")
        except Exception:
            logging.getLogger("HWR").error("error sending message: " + str(msg))


def energy_scan_finished(pk, ip, rm, sample):
    server.emit(
        "energy_scan_result",
        {"pk": pk, "ip": ip, "rm": rm},
        namespace="/hwr",
    )


def queue_interleaved_started():
    node = last_queue_node()

    msg = {
        "Signal": "queue_interleaved_started",
        "Message": "Interleaved collection started",
        "taskIndex": node["idx"],
        "queueID": node["queue_id"],
        "sample": node["sample"],
        "state": RUNNING,
        "progress": 0,
    }

    logging.getLogger("HWR").debug("[TASK CALLBACK] " + str(msg))

    try:
        server.emit("task", msg, namespace="/hwr")
    except Exception:
        logging.getLogger("HWR").error("error sending message: " + str(msg))


def queue_interleaved_finished():
    node = last_queue_node()

    msg = {
        "Signal": "queue_interleaved_finished",
        "Message": "Interleaved collection ended",
        "taskIndex": node["idx"],
        "queueID": node["queue_id"],
        "sample": node["sample"],
        "state": COLLECTED,
        "progress": 1,
    }

    logging.getLogger("HWR").debug("[TASK CALLBACK] " + str(msg))

    try:
        server.emit("task", msg, namespace="/hwr")
    except Exception:
        logging.getLogger("HWR").error("error sending message: " + str(msg))


def queue_interleaved_sw_done(data):
    node = last_queue_node()
    progress = mxcube.queue.get_task_progress(node["node"], data)

    msg = {
        "Signal": "collectImageTaken",
        "Message": task_signals["collectImageTaken"],
        "taskIndex": node["idx"],
        "queueID": node["queue_id"],
        "sample": node["sample"],
        "state": RUNNING if progress < 1 else COLLECTED,
        "progress": progress,
    }

    logging.getLogger("HWR").debug("[TASK CALLBACK] " + str(msg))

    try:
        server.emit("task", msg, namespace="/hwr")
    except Exception:
        logging.getLogger("HWR").error("error sending message: " + str(msg))


def xrf_task_progress(taskId, progress):
    node = last_queue_node()

    msg = {
        "Signal": "XRFTaskUpdate",
        "Message": "XRFTaskUpdate",
        "taskIndex": node["idx"],
        "queueID": node["queue_id"],
        "sample": node["sample"],
        "state": RUNNING if progress < 1 else COLLECTED,
        "progress": progress,
    }

    try:
        server.emit("task", msg, namespace="/hwr")
    except Exception:
        logging.getLogger("HWR").error("error sending message: " + str(msg))


def motor_position_callback(movable):
    server.emit("motor_position", movable, namespace="/hwr")


def new_plot(plot_info):
    try:
        server.emit("new_plot", plot_info, namespace="/hwr")
    except Exception:
        logging.getLogger("HWR").error("error sending new_plot message: %s", plot_info)


@RateLimited(1)
def plot_data(data, last_index=[0], **kwargs):
    data_data = data["data"]
    if last_index[0] > len(data_data):
        last_index = [0]

    data["data"] = data_data[last_index[0] :]

    try:
        server.emit("plot_data", data, namespace="/hwr")
    except Exception:
        logging.getLogger("HWR").exception(
            "error sending plot_data message for plot %s", data["id"]
        )
    else:
        last_index[0] += len(data_data)


def plot_end(data):
    try:
        server.emit("plot_end", data, namespace="/hwr")
    except Exception:
        logging.getLogger("HWR").error(
            "error sending plot_end message for plot %s", data["id"]
        )



import logging
import os
import traceback
from pathlib import Path

import flask_security
import gevent
import werkzeug
from flask import (
    Flask,
    request,
)
from flask_login import current_user
from flask_socketio import SocketIO
from werkzeug.middleware.proxy_fix import ProxyFix

from mxcubeweb.core.components.user.database import (
    UserDatastore,
    init_db,
)
from mxcubeweb.core.models.usermodels import (
    Message,
    Role,
    User,
)
from mxcubeweb.core.server.csp import CSPMiddleware
from mxcubeweb.core.server.limiter import init_limiter
from mxcubeweb.core.server.resource_handler import ResourceHandlerFactory
from mxcubeweb.core.util import networkutils


class Server:
    init_event = gevent.event.Event()
    flask = None
    security = None
    api = None
    user_datastore = None
    db_session = None
    flask_socketio = None
    limiter = None

    def __init__(self):
        msg = "Server is to be used as a pure static class, don't instantiate."
        raise NotImplementedError(msg)

    @staticmethod
    def exception_handler(e):
        err_msg = "Uncaught exception while calling %s" % request.path
        logging.getLogger("exceptions").exception(err_msg)
        return err_msg + ": " + traceback.format_exc(), 409

    @staticmethod
    def init(cmdline_options, cfg):
        template_dir = os.path.join(os.path.dirname(__file__), "templates")

        Server.flask = Flask(
            __name__,
            static_folder=cmdline_options.static_folder,
            static_url_path="",
            template_folder=template_dir,
        )
        Server.flask.wsgi_app = ProxyFix(Server.flask.wsgi_app)
        Server.flask.config.from_object(cfg.flask)

        Server.flask.register_error_handler(Exception, Server.exception_handler)

        if cfg.flask.CSP_ENABLED:
            Server.flask.wsgi_app = CSPMiddleware(
                Server.flask.wsgi_app,
                {
                    "CSP_ENABLED": cfg.flask.CSP_ENABLED,
                    "CSP_POLICY": cfg.flask.CSP_POLICY,
                    "CSP_REPORT_ONLY": cfg.flask.CSP_REPORT_ONLY,
                    "CSP_REPORT_URI": cfg.flask.CSP_REPORT_URI,
                },
            )

        db_session = init_db(cfg.flask.USER_DB_PATH)
        Server.user_datastore = UserDatastore(
            db_session, User, Role, message_model=Message
        )

        Server.db_session = db_session

        Server.flask_socketio = SocketIO(
            manage_session=False,
            cors_allowed_origins=cfg.flask.ALLOWED_CORS_ORIGINS,
        )
        Server.flask_socketio.init_app(Server.flask)

        if cfg.flask.RATE_LIMITER_ENABLED:
            Server.limiter = init_limiter(Server.flask)
        else:
            Server.limiter = None

        # the following test prevents Flask from initializing twice
        # (because of the Reloader)
        if not Server.flask.debug or os.environ.get("WERKZEUG_RUN_MAIN") == "true":
            # Make the valid_login_only decorator available on server object
            Server.restrict = staticmethod(networkutils.auth_required)
            Server.require_control = staticmethod(networkutils.require_control)
            Server.ws_restrict = staticmethod(networkutils.ws_valid_login_only)
            Server.route = staticmethod(Server.flask.route)

    def _register_route(init_blueprint_fn, app, url_prefix):
        bp = init_blueprint_fn(app, Server, url_prefix)
        Server.flask.register_blueprint(bp)

        for key, function in Server.flask.view_functions.items():
            if key.startswith(bp.name) and not hasattr(function, "tags"):
                function.tags = [bp.name.title().replace("_", " ")]

        return bp

    @staticmethod
    def register_routes(mxcube):
        Server.security = flask_security.Security(
            Server.flask, Server.user_datastore, register_blueprint=False
        )

        from mxcubeweb.routes.csp_report import init_route as init_csp_route
        from mxcubeweb.routes.harvester import init_route as init_harvester_route
        from mxcubeweb.routes.login import init_route as init_login_route
        from mxcubeweb.routes.main import init_route as init_main_route
        from mxcubeweb.routes.queue import init_route as init_queue_route
        from mxcubeweb.routes.ra import init_route as init_ra_route
        from mxcubeweb.routes.workflow import init_route as init_workflow_route

        url_root_prefix = "/mxcube/api/v0.1"

        Server._register_route(init_csp_route, mxcube, f"{url_root_prefix}/csp")

        Server._register_route(init_login_route, mxcube, f"{url_root_prefix}/login")

        Server._register_route(init_main_route, mxcube, f"{url_root_prefix}")

        Server._register_route(init_queue_route, mxcube, f"{url_root_prefix}/queue")

        Server._register_route(init_ra_route, mxcube, f"{url_root_prefix}/ra")

        Server._register_route(
            init_workflow_route, mxcube, f"{url_root_prefix}/workflow"
        )

        Server._register_route(
            init_harvester_route, mxcube, f"{url_root_prefix}/harvester"
        )

        ResourceHandlerFactory.register_with_server(Server.flask)

    @staticmethod
    def emit(*args, **kwargs):
        if current_user and current_user.is_authenticated:
            logging.getLogger("server_access").debug(
                f"{current_user.username} websocket emit: {args} {kwargs}"
            )
        Server.flask_socketio.emit(*args, **kwargs)

    @staticmethod
    def run(cfg):
        if cfg.flask.CERT == "SIGNED" and cfg.flask.CERT_PEM and cfg.flask.CERT_KEY:
            ssl_context = werkzeug.serving.load_ssl_context(
                cfg.flask.CERT_PEM, cfg.flask.CERT_KEY
            )
        elif cfg.flask.CERT == "ADHOC":
            cert_dir = Path(cfg.flask.USER_DB_PATH).parent
            ssl_context = werkzeug.serving.load_ssl_context(
                *werkzeug.serving.make_ssl_devcert(str(cert_dir))
            )
        else:
            ssl_context = None

        if ssl_context:
            Server.flask_socketio.run(
                Server.flask,
                ssl_context=ssl_context,
                host=cfg.flask.HOST,
                port=cfg.flask.PORT,
            )
        else:
            Server.flask_socketio.run(Server.flask, cfg.flask.HOST, port=cfg.flask.PORT)



import contextlib
import inspect
import logging
import traceback
import typing
from typing import Any, ClassVar

import gevent
from pydantic import (
    Field,
    ValidationError,
    create_model,
)

from mxcubeweb.core.models.adaptermodels import (
    HOActuatorModel,
    HOModel,
)
from mxcubeweb.core.models.configmodels import ResourceHandlerConfigModel
from mxcubeweb.core.server.resource_handler import (
    ResourceHandlerFactory,
)

default_resource_handler_config = ResourceHandlerConfigModel(
    commands=[
        "set_value",
    ],
    attributes=[
        "data",
        "get_value",
    ],
)


class AdapterBase:
    """Hardware Object Adapter Base class."""

    # List of supported HO base classes (or callable for more advanced matching)
    SUPPORTED_TYPES: ClassVar[list[object]] = []
    # Dictionary of supported types and their adapters
    SUPPORTED_TYPES_TO_ADAPTERS = {}

    ATTRIBUTES = []
    METHODS = []

    ADAPTER_DICT = {}

    def __init__(self, ho, role, app, resource_handler_config=None):  # noqa: D417
        """Initialize.

        Args:
            ho (object): Hardware object to mediate for.
            (str): The name of the object
        """
        self.app = app
        self._ho = ho
        self._name = role
        self._available = True
        self._read_only = False
        self._type = type(self).__name__.replace("Adapter", "").upper()
        self._unique = True
        self._msg = ""

        cls_name = self.__class__.__name__.lower()

        if resource_handler_config:
            cls_name = resource_handler_config.name or cls_name

        if cls_name not in self.ADAPTER_DICT:
            self.ADAPTER_DICT[cls_name] = {}

        if ho is not None:
            self.ADAPTER_DICT[cls_name][ho.name] = self

        if resource_handler_config:
            _name = resource_handler_config.name or cls_name

            ResourceHandlerFactory.create_or_get(
                name=_name,
                url_prefix="/mxcube/api/v0.1/hwobj/"
                + (resource_handler_config.name or self._type.lower()),
                handler_dict=self.ADAPTER_DICT[cls_name],
                app=self.app,
                exports=resource_handler_config.exports,
                commands=resource_handler_config.commands,
                attributes=resource_handler_config.attributes,
            )

    @classmethod
    def can_adapt(cls, ho):
        return any(isinstance(ho, t) for t in cls.SUPPORTED_TYPES)

    def __init_subclass__(cls, **kwargs):
        for ho_class in cls.SUPPORTED_TYPES:
            if ho_class in AdapterBase.SUPPORTED_TYPES_TO_ADAPTERS:
                error_msg = ("Adapter for class %s already exists", ho_class.__name__)
                raise ValueError(error_msg)
            AdapterBase.SUPPORTED_TYPES_TO_ADAPTERS[ho_class] = cls
        super().__init_subclass__(**kwargs)

    @classmethod
    def get_resource_handler(cls):
        return AdapterBase.RESOURCE_HANDLER_DICT.get(cls.__name__, None)

    def get_adapter_id(self, ho=None):
        ho = ho if ho else self._ho
        return self.app.mxcubecore._get_adapter_id(ho)

    def _add_adapter(self, attr_name, ho, adapter_cls):
        _id = f"{self.get_adapter_id()}.{attr_name}"
        adapter_instance = adapter_cls(ho, _id, self.app)
        self.app.mxcubecore._add_adapter(_id, adapter_cls, ho, adapter_instance)

        setattr(self, attr_name, adapter_instance)

    def execute_command(self, cmd_name, args):
        try:
            self._pydantic_model_for_command(cmd_name).validate(args)
        except ValidationError:
            logging.getLogger("MX3.HWR").exception(
                f"Error when validating input {args} for command {cmd_name}"
            )

        task = gevent.spawn(self._execute_command, cmd_name, args)
        task.call_args = {"cmd_name": cmd_name, "value": args}
        task.link_value(self._command_success)

    def _execute_command(self, cmd_name, args):
        _cmd = getattr(self, cmd_name, None)
        logging.getLogger("MX3.HWR").info(
            f"Calling {self._name}.{cmd_name} with {args}"
        )

        try:
            if _cmd:
                return _cmd(**args)
            return self._ho.execute_exported_command(cmd_name, args)
        except Exception as ex:
            self._command_exception(cmd_name, ex)
            logging.getLogger("MX3.HWR").exception("")

    def _command_success(self, t):
        value = t.value
        cmd_name = t.call_args["cmd_name"]
        attr = getattr(self._ho, cmd_name)
        model = self._model_from_typehint(attr)

        try:
            model["return"].validate({"return": value})
        except ValidationError:
            attr_name = t.call_args["cmd_name"]
            logging.getLogger("MX3.HWR").exception(
                f"Return value of {self._name}.{attr_name} is of wrong type"
            )
        else:
            logging.getLogger("MX3.HWR").info(
                f"{self._name}.{cmd_name} returned {value}"
            )
            if value:
                self._msg = value
            self.app.server.emit(
                "hardware_object_command_return",
                {"cmd_name": cmd_name, "value": value},
                namespace="/hwr",
            )

            self.emit_ho_changed(self.state())

    def _command_exception(self, cmd_name, ex):
        self._msg = traceback.format_exc()
        self.app.server.emit(
            "hardware_object_command_error",
            {"cmd_name": cmd_name, "value": str(ex)},
            namespace="/hwr",
        )
        self.emit_ho_changed(self.state())

    @property
    def adapter_type(self):
        """Adapter type.

        Returns:
            (str): The data type of the value
        """
        return self._type

    @property
    def ho(self):
        """Underlaying HardwareObject.

        Returns:
            (object): HardwareObject
        """
        return self._ho

    # Abstract method
    def state(self):
        """Retrieve the state of the underlying hardware object as a JavaScript string.

        Retrieves the state of the underlying hardware object and converts it to a str
        that can be used by the javascript front end.

        Returns:
            (str): The state
        """
        return self._ho.get_state().name

    # Abstract method
    def msg(self):
        """Return a message describing the current state.

        Should be used to communicate details of the state to the user.

        Returns:
            (str): The message string.
        """
        return self._msg

    def read_only(self):
        """Return true if the hardware object is read only.

        Return true if the hardware object is read only and
        ``set_value`` can not be called.

        Returns:
            (bool): True if read enly.
        """
        return self._read_only

    def available(self):
        """Check if the hardware object is considered to be available/online/enabled.

        Returns:
            (bool): True if available.
        """
        return self._available

    def _model_from_typehint(self, attr):
        input_dict = {}
        output_dict = {}

        for _n, _t in typing.get_type_hints(attr).items():
            if _n != "return":
                input_dict[_n] = (_t, Field(alias=_n))
            else:
                if not inspect.isclass(_t):
                    _t = _t.__class__

                output_dict[_n] = (_t, Field(alias=_n))

        return {
            "args": create_model(attr.__name__, **input_dict),
            "return": create_model(attr.__name__, **output_dict),
            "signature": list(input_dict.keys()),
        }

    def _pydantic_model_for_command(self, cmd_name):
        if cmd_name in self.METHODS:
            return self._model_from_typehint(getattr(self, cmd_name, None))["args"]
        return self._ho.pydantic_model[cmd_name]

    def _exported_methods(self):
        exported_methods = {}
        # Get exported attributes from underlaying HardwareObject
        # and only set display True for those methods that have
        # been explicitly configured to be exported. The method also
        # needs to be defined as a command in the ResourceHandlerConfigModel
        # to be exported.

        configured_exported = self._ho.exported_attributes.keys()

        rh = ResourceHandlerFactory.get_handler(self.__class__.__name__.lower())

        if rh:
            for export in rh.commands:
                attr = getattr(self, export["attr"], None)

                if inspect.ismethod(attr):
                    model = self._model_from_typehint(attr)
                    exported_methods[export["attr"]] = {
                        "signature": model["signature"],
                        "schema": model["args"].schema_json(),
                        "display": export["attr"] in configured_exported,
                    }

        return exported_methods

    def commands(self):
        return self._exported_methods()

    def attributes(self):
        _attributes = {}

        for attribute_name in self.ATTRIBUTES:
            attr = getattr(self, attribute_name, None)

            if attr:
                model = self._model_from_typehint(attr)
                value = attr()

                try:
                    model["return"].validate({"return": value})
                except ValidationError:
                    logging.getLogger("MX3.HWR").exception(
                        "Return value of"
                        f" {self._name}.{attribute_name} is of wrong"
                        " type"
                    )
                    _attributes[attribute_name] = {}
                else:
                    _attributes[attribute_name] = attr()

        return _attributes

    def emit_ho_attribute_changed(
        self, attribute: str, value: Any, operation: str = "SET"
    ):
        self.app.server.emit(
            "hardware_object_attribute_changed",
            {
                "name": self._name,
                "attribute": attribute,
                "value": value,
                "operation": operation.upper(),
            },
            namespace="/hwr",
        )

    def emit_ho_value_changed(self, value: Any):
        self.app.server.emit(
            "hardware_object_value_changed",
            {"name": self._name, "value": value},
            namespace="/hwr",
        )

    def emit_ho_changed(self, state, **kwargs):
        """Signal handler to send entire object to the client via socketIO."""
        data = self.data().dict()

        if hasattr(state, "name"):
            data["state"] = state.name
        else:
            logging.getLogger("MX3.HWR").info(
                f"emit_ho_changed with {state} for {self._ho.name}"
            )

        self.app.server.emit("hardware_object_changed", data, namespace="/hwr")

    def state_change(self, state, **kwargs):
        """Signal handler to send the state to the client via socketIO."""
        self.emit_ho_changed(state)

    def _dict_repr(self):
        """Dictionary representation of the hardware object.

        Returns:
            (dict): The dictionary.
        """
        try:
            data = {
                "name": self._name,
                "state": self.state(),
                "msg": self.msg(),
                "type": self._type,
                "available": self.available(),
                "readonly": self.read_only(),
                "commands": self.commands(),
                "attributes": self.attributes(),
            }

        except Exception as ex:
            # Return a default representation if there is a problem retrieving
            # any of the attributes
            self._available = False

            data = {
                "name": self._name,
                "state": "UNKNOWN",
                "msg": "Exception: %s" % str(ex),
                "type": "FLOAT",
                "available": self.available(),
                "readonly": False,
                "commands": {},
                "attributes": {},
            }

            logging.getLogger("MX3.HWR").exception(
                f"Failed to get dictionary representation of {self._name}"
            )
        return data

    def data(self) -> HOModel:
        return HOModel(**self._dict_repr())


class ActuatorAdapterBase(AdapterBase):
    def __init__(self, ho, role, app, resource_handler_config=None):  # noqa: D417
        """Initialize.

        Args:
            (object): Hardware object to mediate for.
            (str): The name of the object.
        """
        super().__init__(ho, role, app, resource_handler_config)

        self._unique = False

        with contextlib.suppress(AttributeError):
            self._read_only = ho.read_only

    # Don't limit rate this method with utils.LimitRate, all subclasses
    # will share this method thus all methods will be effected if limit rated.
    # Rather LimitRate the function calling this one.
    def value_change(self, *args, **kwargs):
        """Signal handler to send values to the client via socketIO."""
        self.emit_ho_value_changed(args[0])

    # Abstract method
    def set_value(self, value) -> str:
        """Sets a value on underlying hardware object.

        Args:
            value(float): Value to be set.

        Returns:
            (str): The actual value set as str.

        Raises:
            ValueError: When conversion or treatment of value fails.
            StopIteration: When a value change was interrupted (abort/cancel).

        Emits:
            hardware_object_value_changed with values over websocket
        """

    # Abstract method
    def get_value(self):
        """Retrieve value from underlying hardware object.

        Returns:
            (str): The value.

        Raises:
            ValueError: When value for any reason can't be retrieved.
        """
        return self._get_value().value

    # Abstract method
    def stop(self):
        """Stop an action/movement."""

    def limits(self):
        """Read the energy limits.

        Returns:
            (tuple): Two floats (min, max).

        Raises:
            ValueError: When limits for any reason can't be retrieved.
        """
        try:
            # Limits are None when not configured, convert them to -1, -1
            # as we are returning floats
            return (0, 0) if None in self._ho.get_limits() else self._ho.get_limits()
        except (AttributeError, TypeError):
            msg = "Could not get limits"
            raise ValueError(msg)

    def _dict_repr(self):
        """Dictionary representation of the hardware object.

        Returns:
            (dict): The dictionary.
        """
        data = super()._dict_repr()

        try:
            data.update({"value": self.get_value().value, "limits": self.limits()})
        except Exception as ex:
            logging.getLogger("MX3.HWR").exception(
                f"Could not get dictionary representation of {self._ho.name}"
            )
            logging.getLogger("MX3.HWR").error(
                f"Check status of {self._ho.name}, object is"
                " offline, in fault or returns unexpected value !"
            )

            self._available = False
            data.update(
                {
                    "value": 0,
                    "limits": (0, 0),
                    "type": "FLOAT",
                    "msg": "Exception %s" % str(ex),
                }
            )

        return data

    def data(self) -> HOActuatorModel:
        return HOActuatorModel(**self._dict_repr())



from mxcubeweb.core.components.user.usermanager import BaseUserManager


class DummyUserManager(BaseUserManager):
    def __init__(self, app, config):
        super().__init__(app, config)

    def _login(self, login_id, password):
        # Assuming that ISPyB is used
        return self.app.lims.lims_login(login_id, password, create_session=False)

    def _signout(self):
        pass



import logging

from flask import (
    Blueprint,
    jsonify,
    make_response,
    redirect,
    request,
    session,
    url_for,
)
from flask_login import current_user

from mxcubeweb.core.util import networkutils


def init_route(app, server, url_prefix):
    bp = Blueprint("login", __name__, url_prefix=url_prefix)

    @bp.route("/", methods=["POST"])
    def login():
        """Login into mxcube application.

        :returns: Response Object, Content-Type: application/json, an object
        containing following info:

        {'status':{ 'code': 'ok', 'msg': msg },
            'Proposal': proposal,
            'session': todays_session,
            'local_contact': local_contact,
            'person': someone,
            'laboratory': a_laboratory]}

        Status code set to:
        200: On success
        409: Error, could not log in
        """
        params = request.get_json()
        login_id = params.get("proposal", "")
        password = params.get("password", "")

        try:
            app.usermanager.login(login_id, password)
        except Exception:
            msg = "[LOGIN] User %s could not login" % login_id
            logging.getLogger("MX3.HWR").exception(msg)
            res = make_response(
                jsonify({"msg": "Authentication failed"}),
                200,
            )
        else:
            res = make_response(jsonify({"msg": ""}), 200)

        return res

    @bp.route("/sso_post_logout", methods=["GET"])
    @server.restrict
    def ssosignout():
        app.usermanager.signout()
        return redirect("/")

    @bp.route("/ssologin", methods=["GET"])
    def ssosignin():
        redirect_uri = url_for("login.auth", _external=True)
        return app.usermanager.oauth_client.keycloak.authorize_redirect(redirect_uri)

    @bp.route("/auth", methods=["GET"])
    def auth():
        try:
            app.usermanager.sso_validate()
        except Exception:
            return redirect("/login")
        else:
            return redirect("/datacollection")

    @bp.route("/signout")
    @server.restrict
    def signout():
        """Signout from MXCuBE Web and reset the session."""
        app.usermanager.signout()
        return make_response(jsonify(""), 200)

    @bp.route("/login_info", methods=["GET"])
    def login_info():
        """Retrieve session/login info.

        :returns: Response Object, Content-Type: application/json, an object
                containing::

            {'synchrotron_name': synchrotron_name,
            'beamline_name': beamline_name,
            'loginType': loginType,
            'loggedIn': True,
            'Proposal': proposal, 'session': todays_session,
            'local_contact': local_contact,
            'person': someone,
            'laboratory': a_laboratory']}}

        Status code set to:
        200: On success
        200: Error, could not log in, {"loggedIn": False}
        """
        res = app.usermanager.login_info()

        if res["loggedIn"]:
            session.permanent = True

        return jsonify(res)

    @bp.route("/send_feedback", methods=["POST"])
    @server.restrict
    def send_feedback():
        sender_data = request.get_json()
        sender_data["LOGGED_IN_USER"] = current_user.nickname
        networkutils.send_feedback(sender_data)
        return make_response("", 200)

    @bp.route("/refresh_session", methods=["GET"])
    @server.restrict
    def refresh_session():
        # Since default value of `SESSION_REFRESH_EACH_REQUEST` config setting is `True`
        # there is no need to do anything to refresh the session.
        app.usermanager.update_active_users()
        app.usermanager.handle_sso_logout()
        return make_response("", 200)

    return bp
